// ISS Tracker - David Hunt
// Version 1.4
// Licence - Attribution Non-commercial Share Alike (by-nc-sa)

// download all the necessary libraries using the library manager
#include <AccelStepper.h> // to drive stepper motors
#include <ezTime.h>
#include <WiFi.h>
#include <HTTPClient.h> // to retrieve data from API
#include <ArduinoJson.h> // to extract ISS location information
#include <U8g2lib.h> // search for U8g2 in library manager
#include <Wire.h>

time_t next_pass_timestamp = 1572652800; // 2 Nov 2019 as a reference point

const char ssid[] = "Your network here";  //  your network SSID (name)
const char pass[] = "Router password";       // your network password
const int touch_threshold = 40;  // touchpin input (T0) is GPIO4
const float steps_per_degree = 4096 / 360; //4096 steps per revolution in half step mode
double my_latitude = 51.89970; // change to your location here - this is used to find Polaris too
double my_longitude = -2.12084; // find your co-ordinates using google maps

String next_pass_JSON; // location data is retrieved in JSON format
int next_risetime;
int next_duration;

U8G2_SSD1306_128X32_UNIVISION_F_HW_I2C u8g2(U8G2_R0);
// Motor pin definitions in GPIO numbers
#define motor_pin_az1  13     // IN1 on the ULN2003 driver
#define motor_pin_az2  12     // IN2 on the ULN2003 driver
#define motor_pin_az3  14     // IN3 on the ULN2003 driver
#define motor_pin_az4  27     // IN4 on the ULN2003 driver

#define motor_pin_el1  26     // IN1 on the ULN2003 driver
#define motor_pin_el2  25     // IN2 on the ULN2003 driver
#define motor_pin_el3  33     // IN3 on the ULN2003 driver
#define motor_pin_el4  32     // IN4 on the ULN2003 driver

#define MotorInterfaceType 8 // 8 = half step mode, 4 = full step mode
// Initialize with pin sequence IN1-IN3-IN2-IN4 for using the AccelStepper library with 28BYJ-48 stepper motor:
AccelStepper azimuth_motor = AccelStepper(MotorInterfaceType, motor_pin_az1, motor_pin_az3, motor_pin_az2, motor_pin_az4);
AccelStepper elevation_motor = AccelStepper(MotorInterfaceType, motor_pin_el1, motor_pin_el3, motor_pin_el2, motor_pin_el4);

Timezone my_timezone;

// use the internet to look up the next flyby time
void get_next_pass(double lat, double lon, int &p_risetime, int &p_duration){
  HTTPClient http;
  const char* api_host = "http://api.open-notify.org/iss-pass.json?lat=";
  char my_lat_position[10];
  char my_lon_position[10];
  char api_address[255];
  // convert the coordinates stored as doubles to strings with 4 dp precision
  dtostrf(lat,7,4,my_lat_position);
  dtostrf(lon,7,4,my_lon_position);
  // create the url with get request
  strcpy(api_address, api_host);
  strcat(api_address, my_lat_position);
  strcat(api_address, "&lon=");
  strcat(api_address, my_lon_position);
  // only get the time and date of the next pass
  strcat(api_address, "&n=1");  
  
  http.begin(api_address);
  int httpCode = http.GET();   
  if (httpCode > 0) {
    String payload = http.getString();
    // Serial.println(httpCode);
    // Serial.println(payload);
    const size_t capacity = JSON_OBJECT_SIZE(2) + JSON_OBJECT_SIZE(3) + 200;
    DynamicJsonDocument doc(capacity);
    DeserializationError error = deserializeJson(doc, payload);
    p_duration = doc["response"][0]["duration"];// write this value to next_duration
    p_risetime = doc["response"][0]["risetime"];// write this value to next_risetime
  }
  else {
    Serial.println("Error on HTTP request");
  }
  http.end(); // Free the resources
}
// convert this to human readable form in days, hours, minutes and seconds
String time_to_next_flyby(int next_observation){
  int time_remaining;
  int next_rise_days;
  int next_rise_hours;
  int next_rise_minutes;
  int next_rise_seconds;
  String countdown_time;
  time_remaining = next_observation - now();
  next_rise_days = time_remaining/(3600 * 24);
  time_remaining %= (3600 * 24);
  next_rise_hours = time_remaining / 3600;
  time_remaining %= 3600;
  next_rise_minutes = time_remaining / 60;
  next_rise_seconds = time_remaining % 60;
  countdown_time += next_rise_days;
  countdown_time += " days ";
  countdown_time += next_rise_hours;
  
  if (next_rise_minutes < 10){
    countdown_time += ":0";}// add leading zero for single digit
  else{
    countdown_time += ":";
  }
  countdown_time += next_rise_minutes;
  if (next_rise_seconds < 10){
    countdown_time += ":0";}// add leading zero for single digit
  else{
    countdown_time += ":";
  }
  countdown_time += next_rise_seconds;
  return countdown_time;
}

// find the current location of the ISS over the Earth
// and calculate the angles in order to point to it from our location
void point_to_ISS(){
  HTTPClient http;
  double latitude;
  double longitude;
  if (http.begin("http://api.open-notify.org/iss-now.json")){
    int httpCode = http.GET();
    // httpCode will be negative if there is an error   
    if (httpCode > 0) {
      String ISS_Location_Data = http.getString();
      //Serial.println(httpCode); // this will be 200 for a successful retrieval
      //Serial.println(ISS_Location_Data); // Json data
      const size_t capacity = JSON_OBJECT_SIZE(2) + JSON_OBJECT_SIZE(3) + 200;
      DynamicJsonDocument doc(capacity);
      DeserializationError error = deserializeJson(doc, ISS_Location_Data);
      //const char* message = doc["message"];
      long timestamp = doc["timestamp"];
      latitude = doc["iss_position"]["latitude"];
      longitude = doc["iss_position"]["longitude"];
      
      // Print values useful for debugging to check it is working
      //Serial.println(timestamp);
      //Serial.println(latitude, 4);
      //Serial.println(longitude, 4);
    }
    else {
      Serial.print("HTTP error :");
      Serial.println(http.errorToString(httpCode).c_str());
    }
    http.end(); // Free the resources
  }
  else {
    Serial.print("HTTP error: unable to connect");
  }
  
  // ESP32 calculates angles in radians (not degrees)
  double latitude_in_radians = latitude * PI / 180;
  double longitude_in_radians = longitude * PI / 180;
  double my_latitude_in_radians = my_latitude * PI / 180;
  double my_longitude_in_radians = my_longitude * PI / 180;
  // calculate azimuth bearing to ISS
  double x = cos(latitude_in_radians)*sin(longitude_in_radians-my_longitude_in_radians);
  double y = cos(my_latitude_in_radians)*sin(latitude_in_radians
             -sin(my_latitude_in_radians*cos(latitude_in_radians
             *cos(longitude_in_radians-my_longitude_in_radians);
  double b = atan2(x,y);
  //Serial.print("Bearing to ISS: ");
  b=b*180/PI;// convert to degrees
  //Serial.println(b);

  // calculate elevation bearing to ISS
  double earth_radius = 6371; // mean value in km
  double orbit = 412; // mean value in km - perhaps look this value up too?
  double x1 = earth_radius * cos(my_latitude_in_radians) * cos(my_longitude_in_radians);
  double y1 = earth_radius * cos(my_latitude_in_radians) * sin(my_longitude_in_radians);
  double z1 = earth_radius * sin(my_latitude_in_radians);
  double x2 = earth_radius * cos(latitude_in_radians) * cos(longitude_in_radians);
  double y2 = earth_radius * cos(latitude_in_radians) * sin(longitude_in_radians);
  double z2 = earth_radius * sin(latitude_in_radians);
  
  double dist_to_nadir = sqrt(sq(x2-x1)+sq(y2-y1)+sq(z2-z1));
  //Serial.print("Distance to nadir :");
  //Serial.println(dist_to_nadir);
  double geocentric_angle_in_radians = 2 * asin(dist_to_nadir/(2*earth_radius));
  //Serial.print("Geocentric angle in degrees :");
  //Serial.println(geocentric_angle_in_radians * 180 / PI);
  double distance_to_ISS = sqrt(sq(orbit+earth_radius)+sq(earth_radius)-2*(orbit+earth_radius)*earth_radius*cos(geocentric_angle_in_radians));
  //Serial.print("Distance to ISS :");
  //Serial.println(distance_to_ISS);
  double ISS_angle = (180 / PI) * asin((orbit + earth_radius) * (sin(geocentric_angle_in_radians)/distance_to_ISS));
  
  // azimuth stepper motor has 4096 half steps per revolution and gear ratio 113/22
  azimuth_motor.moveTo(4096/360 * 113/22 * b);
  // elevation motor has 4096 half steps per revolution
  elevation_motor.moveTo(4096/ 360 * ISS_angle); 
    
  // run the motor whilst waiting to update the co-ordinates every 5 seconds
  double holding_time = now();
  while (now()-holding_time < 5){
    azimuth_motor.run();
    elevation_motor.run();
    yield();
  }
}

// print out messages on OLED using two lines of text
void display_message(String line_one, String line_two){
  u8g2.setFont(u8g2_font_8x13_tr);
  //u8g2.setFont(u8g2_font_fur14_tr);
  u8g2.firstPage();
  u8g2.setCursor(0, 15);
  u8g2.print(line_one);
  u8g2.setCursor(0, 31);
  u8g2.print(line_two);
  u8g2.nextPage();
}

void setup() {
  Serial.begin(115200);
  delay(3000); // try this to allow OLED to establish communication?
  u8g2.begin(); //for OLED
  u8g2.enableUTF8Print();//
  // Set the maximum motor steps per second:
  azimuth_motor.setMaxSpeed(800); // large gear ratio, so speed this up
  azimuth_motor.setAcceleration(100.0);
  elevation_motor.setMaxSpeed(400);
  elevation_motor.setAcceleration(100.0);

  // ######################### calibration section ############################### 
  display_message("Stop arrow when","pointing down");
  while (touchRead(T0)> touch_threshold){
    elevation_motor.move(400);
    elevation_motor.run();
  }
  // button has been pressed when elevation pointer is downwards
  display_message("Elevation motor","set to zero");
  elevation_motor.setCurrentPosition(0);
  delay(1000);
  display_message("Stop arrow on","left hand side");
  while (touchRead(T0)> touch_threshold){
    azimuth_motor.move(400);
    azimuth_motor.run();
  }
  // button is pressed when azimuth pointer is forwards
  display_message("Azimuth motor","set to zero");
  delay(2000); //pause to let you know button has been pressed
  azimuth_motor.setCurrentPosition(0);
  // move the elevation to point to Polaris
  // this angle is the same as the latitude of the observer, reference point downwards, so +90
  elevation_motor.moveTo(steps_per_degree * (my_latitude + 90));
  elevation_motor.runToPosition();
  display_message("Point to Polaris","Then hit button");
  while (touchRead(T0)> touch_threshold){}// wait for button press
  display_message("Now calibrated","");
  delay(2000); //pause to let you know button has been pressed
  // ######################### end of calibration section ############################ 
  
  // ######################### Connect to WiFi #######################################
  Serial.println("ISS Tracker");
  display_message("Searching for",ssid);
  Serial.println(ssid);
  WiFi.begin(ssid, pass);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("Connected to wifi");
  display_message("Connected to",ssid);
  
  // ######################### Synchronise Time with UTC ###############################
  // Uncomment the line below to see what it does behind the scenes
  setDebug(INFO);
  waitForSync(); // don't do anything until it has the correct time
  
  Serial.println("Current UTC time: " + UTC.dateTime());
  my_timezone.setLocation(F("GB"));// https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
  delay(3000);
  Serial.print(F("Local Time: "));
  Serial.println(my_timezone.dateTime());
}

void loop() {
  events();// this ensures the motors keep moving
  //Serial.println("COOKIE:      " + UTC.dateTime(COOKIE));
  //  Serial.println(dateTime(next_pass_timestamp, "D d-M H:i:s"));
  if (now()+next_duration > next_risetime){// only get next flyby once current one has gone by
    get_next_pass(my_latitude,my_longitude,next_risetime,next_duration);  
  }
  Serial.print("Next flyby:");
  //Serial.print(dateTime(next_risetime, "D d-M H:i:s  "));
  Serial.println(time_to_next_flyby(next_risetime));
  point_to_ISS();
  delay(1000);
  //display_message(UTC.dateTime("H:i:s T"),my_timezone.dateTime("H:i:s"));
  display_message(UTC.dateTime("H:i:s T"),time_to_next_flyby(next_risetime));
}
